/********************************
* This file is part of HFApplet, copyright (C) 2005 Simon Rochester and John C. Morrison
*
* HFApplet is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* HFApplet is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**************************************/

package edu.louisville.physics.jcm.hfapplet;

import java.util.*;

public class ElectronConfiguration implements Cloneable{
//	protected boolean contractClosedShell = true;
	public Term term;
	static int[] nobleNumbers = {86,54,36,18,10};
	ArrayList<ElectronSubshell> configuration;
	
	//	constructor
	public ElectronConfiguration() {
		term = new Term();
		configuration = new ArrayList<ElectronSubshell>();		
	}

	public ElectronConfiguration(int atomicNumber) {
		this();
		setToGroundState(atomicNumber);
	}

	public Object clone() {
		ElectronConfiguration c = new ElectronConfiguration();
		for (int i=0; i<configuration.size(); i++) {
			c.configuration.add((ElectronSubshell)getSubshellAt(i).clone());
		}
		c.setTerm((Term)getTerm().clone());
	//	c.contractClosedShell = contractClosedShell;
		return c;
	}

	public boolean equals(ElectronConfiguration ec) {
//		System.out.println(ec);
//		return true;
		return (configuration.equals(ec.configuration) && term.equals(ec.term));
	}
	

	//	setToGroundState: removes all shells and adds ground state electrons as given in ElementsData
	public void setToGroundState(int atomicNumber) {
		//clear configuration
		configuration.clear();
		term = new Term();

		if (atomicNumber > 0) {
			//get data from ElementsData
			int[][] groundState = ElementsData.getGroundState(atomicNumber);

			//add new electrons, assuming data in ground state is in ascending order
			ElectronSubshell shell;
			for (int i=0; i<groundState.length; i++) {
				shell = new ElectronSubshell(groundState[i][0],groundState[i][1]);
				shell.addElectrons(groundState[i][2]);
				configuration.add(shell);
			}

			term.setToGroundState(atomicNumber);
		}
	}
	
	public int size() {
		return configuration.size();
	}
	
	public int indexOf(ElectronSubshell e) {
		return configuration.indexOf(e);
	}
	
	//setTerm
	public void setTerm(Term term) {
		this.term = term;
	}

	//getTerm
	public Term getTerm() {
		return term;
	}

	// getMultiplicity
	public int getMultiplicity() {
		return term.getMultiplicity();
	}

	// getL
	public int getL() {
		return term.getL();
	}

	// setMultiplicity
	public void setMultiplicity(String multi) {
		term.setMultiplicity(multi);
	}

	// setL
	public void setL(String L) {
		term.setL(L);
	}

	// setContractClosedShell
//	public void setContractClosedShell(boolean b) {
//		contractClosedShell = b;
//	}
	
	// getContractClosedShell
//	public boolean getContractClosedShell() {
//		return contractClosedShell;
//	}

	//	getSubshellAt: typecast a vector element to ElectronSubshell
	public ElectronSubshell getSubshellAt(int i) {
		return configuration.get(i);
	}

	// getSubshell
	public ElectronSubshell getSubshell(int n, int j) throws ShellException {
		return getSubshell(new ElectronSubshell(n,j));
	}

	public ElectronSubshell getSubshell(String nj) throws ShellException {
		return getSubshell(new ElectronSubshell(nj));
	}

	public ElectronSubshell getSubshell(ElectronSubshell shell) throws ShellException {
		int compare;
		for (int i=0; i<configuration.size(); i++) {
			if ((compare = shell.compareTo(getSubshellAt(i))) == 0) {
				return getSubshellAt(i);
			} else if (compare < 0) {
				throw new ShellException(ShellException.EMPTY);
			}
		}
		throw new ShellException(ShellException.EMPTY);
	}

	// addElectron: adds an electron to specified shell.
	//		Creates new shell in correct order if shell does not exist
	public void addElectron(ElectronSubshell shell) throws ShellException {
		ElectronSubshell existingShell = null;
		int i = 0;
		int compare = 1;
		
		try {
			while (true) {
				if (i == configuration.size()) {
					configuration.add(shell);
					shell.addElectron();
					return;
				}
				existingShell = getSubshellAt(i);
				compare = shell.compareTo(existingShell);
				if (compare < 0) {
					configuration.add(i,shell);
					shell.addElectron();
					return;
				};
				if (compare == 0) {
					existingShell.addElectron();
					return;
				}
				i++;
			}
		} catch (ShellException e) {
			throw e;
		}
	}
	
	// removeElectron: removes electron from specified shell. 
	//		Throws exception if shell does not exist or is empty
	public void removeElectron(ElectronSubshell shell) throws ShellException {
		ElectronSubshell existingShell = null;
		int i = 0;
		int compare = 1;
		do {
			if (i == configuration.size()) 
				throw new ShellException(ShellException.EMPTY);
			existingShell = getSubshellAt(i);
			compare = shell.compareTo(existingShell);
			if (compare < 0) {
				throw new ShellException(ShellException.EMPTY);
			};
			if (compare == 0) {
				try {
					existingShell.removeElectron();
					if (existingShell.isEmpty())
						configuration.remove(i);
				} catch (ShellException e) {
					configuration.remove(i);
					throw e;
				};
			};
			i++;
		} while (i <= configuration.size() && compare == 1);
	}

	public String toString() {
		String string = "";
		for (int i=0; i<configuration.size(); i++) {
			string += getSubshellAt(i).toString();
		}
		return string;
	}

	public ElectronConfiguration getOutsideNoble() {
		ElectronConfiguration outsideConfig = (ElectronConfiguration) this.clone();
		ElectronConfiguration nobleConfig = new ElectronConfiguration();
		for (int i=0; i<nobleNumbers.length; i++) {
			nobleConfig.setToGroundState(nobleNumbers[i]);
			if (outsideConfig.containsConfig(nobleConfig) 
				&& !outsideConfig.configuration.equals(nobleConfig.configuration)) {
				outsideConfig.removeShells(nobleConfig);
				break;
			}
		}
		return outsideConfig;
	}

	public boolean containsShell(ElectronSubshell e) {
		return this.configuration.contains(e);
	}

	public boolean containsConfig(ElectronConfiguration c) {
		for (ElectronSubshell e : c.configuration) {
			if (!this.configuration.contains(e)) return false;
		}
		return true;
	}

	public void removeShells(ElectronConfiguration c) {
		for (ElectronSubshell e : c.configuration) {
			this.configuration.remove(e);
		}
	}
		
	
//	public String[] subshellsToShow() {
//		if (contractClosedShell) {
//			return getOutsideNoble().getSubshellStringArray();
//		} else {
//			return getSubshellStringArray();
//		}
//	}

	public String[] getSubshellStringArray() {
		String[] array = new String[configuration.size()];
		for (int i=0; i<configuration.size(); i++) {
			array[i] = getSubshellAt(i).subshellString();
		}
		return array;
	}

	public int getNobleCoreNumber() {
		ElectronConfiguration nobleConfig = new ElectronConfiguration();
		for (int i=0; i<nobleNumbers.length; i++) {
			nobleConfig.setToGroundState(nobleNumbers[i]);
			if (containsConfig(nobleConfig) 
				&& !configuration.equals(nobleConfig.configuration)) return nobleNumbers[i];
		}
		return 0;
	}

	public void removeCore(int atomicNumber) {
		ElectronConfiguration core = new ElectronConfiguration();
		core.setToGroundState(atomicNumber);
		removeShells(core);
	}

/*	public String toHTMLString() {
		String string = "<html>&nbsp;";
		ElectronConfiguration displayConfig = (ElectronConfiguration) this.clone();
		
		if (contractClosedShell) {
			ElectronConfiguration nobleConfig = new ElectronConfiguration();
			for (int i=0; i<nobleNumbers.length; i++) {
				nobleConfig.setToGroundState(nobleNumbers[i]);
				if (displayConfig.containsAll(nobleConfig) && !displayConfig.equals(nobleConfig)) {
					displayConfig.removeAll(nobleConfig);
					string += "[" + ElementsData.getAbbreviation(nobleNumbers[i]) + "]<small> </small>";
					break;
				}
			}
		}

		for (int i=0; i<displayConfig.size(); i++) {
			string += displayConfig.getSubshellAt(i).toHTMLString();
			string += "<small> </small>";
		}

		return string + "&nbsp;" + term.toHTMLString() + "</html>";
	}
*/
	public String closedSubshellString() {
		ElectronSubshell shell = null;
		String string = " ";

		for (int i=0; i<configuration.size(); i++) {
			shell = getSubshellAt(i);
			if (shell.isFull()) {
				string += shell.subshellString();
				string += "  ";
			}
		}
		return string;
	}

	public String openShellString() {
		ElectronSubshell shell = null;
		String string = "";

		for (int i=0; i<configuration.size(); i++) {
			shell = getSubshellAt(i);
			if (!shell.isFull()) 
				string += shell.toString();
		}
		return string;
	}


}
