/********************************
* This file is part of HFApplet, copyright (C) 2005 Simon Rochester and John C. Morrison
*
* HFApplet is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* HFApplet is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**************************************/

package edu.louisville.physics.jcm.hfapplet;

import java.io.*;
import java.net.*;
import javax.swing.*;
import javax.swing.table.*;
import java.text.*;

public class HFIO {
	Socket socket = null;
	PrintWriter writer = null;
	BufferedReader reader = null;
	StreamTokenizer tokenizer;
	boolean debug = false;

	int orbitalCount;
	public double[][] wfArray;
	public double[][] estimatesArray;
	public double[][] expectationArray;
	public String[] nlArray;
	public JTextArea outputText, iterationsText, ouf = null;
	double[] energies;
	double[] slaterValues;
	double rkValue;
	double[] soValues;
	public int iterationCount;

	final static int PORT = 9096;

	public String connect(String host) throws IOException {
		socket = new Socket(host, PORT);
		writer = new PrintWriter(socket.getOutputStream(), true);
        reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));
		tokenizer = new StreamTokenizer(reader);
		return "Connected to " + socket.getInetAddress().getHostName() + ":" + socket.getPort();
	}
		
	//cleanup
	void disconnect() throws IOException {
		if (writer != null) exit();
		if (socket != null) socket.close();
		if (writer != null) writer.close();
		if (reader != null) reader.close();
	}

	public void endCalculation() {
		writeln("RESTART");
	}

	public void exit() {
		writeln("EXIT");
	}

	private void writeln(String s) {
		if (debug) System.out.println("Input:  "+s);
		writer.println(s);
	}

	private String readln() throws IOException{
		String line = reader.readLine();
		if (debug) System.out.println("Output: "+line);
		return line;
	}

	private void nextToken() throws IOException{
		tokenizer.nextToken();
		if (debug) System.out.println(tokenizer.toString());
	}
		

	public void requestData(String elemAbbr, int atomicNumber, ElectronConfiguration config, 
							boolean reuseWF, boolean strong, double acc) throws IOException, HFIOException {
		if (config.size() == 0) throw new HFIOException("no electrons");
		writeln(elemAbbr + ","					//atom
			+ config.getTerm().toString() + ","	//term
			+ atomicNumber);					//Z
		writeln(reuseWF ? "Y" : "N");			//reuse wavefunctions?
		writeln(config.closedSubshellString());	//closed shells
		writeln(config.openShellString());		//electrons outside closed shells
		writeln("ALL");							//orbitals to be varied
		writeln("N");							//default electron parameters?
		writer.println(acc);
		writeln("END");

		//read response and write it to outputText until configuration data begins
		String line;
		outputText = new JTextArea();
		while (!(line = readln()).equals(" ORBITALS")) {
			outputText.append(line + "\n");
		}

		//read configuration data
		nextToken();
		orbitalCount = (int) Math.round(tokenizer.nval);
		nlArray = new String[orbitalCount];
		
		for (int i=0; i<orbitalCount; i++) nlArray[i] = readOrbital(tokenizer);

		//read response from server and write it to outputText until input data begins
		while (!(line = readln()).equals(" INPUT DATA")) { 
			outputText.append(line + "\n");
		}
		readln();

		//read input data
		estimatesArray = new double[3][orbitalCount];

		for (int i=0; i<orbitalCount; i++) {
			readOrbital(tokenizer);
			nextToken();							//read sigma
			estimatesArray[0][i] = tokenizer.nval;
			nextToken();							//read meth
			nextToken();							//read acc
			estimatesArray[1][i] = tokenizer.nval;
			nextToken();							//read function
			estimatesArray[2][i] = tokenizer.nval;
		}
		
	//	readln();
		line = readln();
		if (line.equals(" AMBIGUOUS TERM")) throw new HFIOException("ambiguous term");
		if (line.equals(" UNSUPPORTED CONFIGURATION")) throw new HFIOException("unsupported configuration");

		if (reuseWF) {										//send wavefunction data
			String[] oufArray = ouf.getText().split("\n");
			for (int i=0; i<oufArray.length; i++) writeln(oufArray[i]);
//			for (int i=0; i<4; i++) writeln(oufArray[i]);
			writeln("ENDOUF");
		}
		writeln("N");								//default remaining parameters?
		writeln("N");								//Default values (NO,STRONG)?
		writeln("200 " + strong);					//Enter values in FORMAT(I3,1X,L1)
		writeln("Y");								//Default values for PRINT, SCFTOL ? (Y/N/H)
		writeln("Y");								//Default values for NSCF, IC ? (Y/N/H)
		writeln("Y");								//Default values for TRACE ? (Y/N/H) 
		writeln("Y");								//Additional parameters ? (Y/N/H)
		writeln("END");								//end of data marker

		//read response from server and write it to outputText until iterations begin
		while (!(line = readln()).equals("ITERATION NUMBER  1")) {
			outputText.append(line + "\n");
			if (line.equals(" UNSUPPORTED CONFIGURATION")) throw new HFIOException("unsupported configuration");
			if (line.equals(" Effective nuclear charge zero for")) throw new HFIOException("charge zero", readOrbital(tokenizer));
		}

		//read response from server and write it to interations until wavefunction plot data begins
		iterationsText = new JTextArea();
		iterationsText.append(line + "\n");
		while (!(line = readln()).equals("BEGIN PLOT DATA")) {
			iterationsText.append(line + "\n");
			if (line.equals(" SCF ITERATIONS HAVE CONVERGED TO THE ABOVE ACCURACY")) {
				nextToken();
				iterationCount = toInt(tokenizer.nval);
			//	while (!readln().equals(" END OF CASE"));
				throw new HFIOException("did not converge", String.valueOf(iterationCount));
			}
		}
		
		//get data array dimensions	
		nextToken();
		int wfPoints = toInt(tokenizer.nval);
		nextToken();
		wfArray = new double[orbitalCount+1][wfPoints];

		//read orbital list
		for (int j=0; j<orbitalCount; j++) readOrbital(tokenizer);

		//read plot data into wfArray
		for (int i=0; i<wfPoints; i++) {
			for (int j=0; j<=orbitalCount; j++) {
				do {
					nextToken();
				} while (!(tokenizer.ttype == StreamTokenizer.TT_NUMBER));
				wfArray[j][i] = tokenizer.nval;
			}
		}

		//continue to read response from server until ouf data begins
		while (!(line = readln()).equals(" BEGIN OUF DATA"));

		//read ouf data
		ouf = new JTextArea();
		while (!(line = readln()).equals(" END OUF DATA")) ouf.append(line + "\n");

		//continue to read response from server until energy values begin
		while (!(line = readln()).equals("BEGIN ENERGY VALUES")) {
			if (line.equals(" END OF CASE")) throw new HFIOException("failed");
		}

		//read energy values
		expectationArray = new double[7][orbitalCount];
		
		for (int i=0; i<orbitalCount; i++) {
			readOrbital(tokenizer);
			nextToken();	//E
			expectationArray[4][i] = tokenizer.nval/2;
			nextToken();	//I
			expectationArray[5][i] = tokenizer.nval;
			nextToken();	//KE
			expectationArray[6][i] = tokenizer.nval;
			nextToken();	//rel
			nextToken();	//S
			nextToken();	//Az
		}
		
		//continue to read response from server until position expectations begin
		while (!readln().equals("BEGIN POSITION EXPECTATIONS"));

		//read position expectations
		for (int i=0; i<orbitalCount; i++) {
			readOrbital(tokenizer);
			nextToken();	//deltaR
			nextToken();	//1/(r^3)
			expectationArray[2][i] = tokenizer.nval;
			nextToken();	//1/r
			expectationArray[1][i] = tokenizer.nval;
			nextToken();	//r
			expectationArray[0][i] = tokenizer.nval;
			nextToken();	//r^2
			expectationArray[3][i] = tokenizer.nval;
		}

		//continue to read response from server until total energy data begins
		while (!readln().equals("TOTAL ENERGY (a.u.)"));

		energies = new double[4];

		//read total energy data
		readln();
		nextToken();
		energies[0] = tokenizer.nval;	//total energy (nonrelativistic)
		readln();
		readln();
		readln();
		readln();
		readln();
		nextToken();
		energies[1] = tokenizer.nval;	//kinetic
		readln();
		nextToken();
		energies[2] = tokenizer.nval;	//potential
		readln();
		nextToken();
		energies[3] = tokenizer.nval;	//ratio
		readln();

		//continue to read response from server until END
		while (!(line = readln()).equals("END")) outputText.append(line + "\n");
	}

	public void getRKIntegral(String input) throws IOException {
		writeln("1");
		writeln(input);
		writeln("END");

		String line;
		while (!(line = readln()).equals("VALUE OF")) {
			if (line.equals("NOT FOUND IN ELECTRON LIST")) throw new IOException();
		}
		readln();
		nextToken();
		rkValue = tokenizer.nval;
	}

	public void getSlaterIntegral(String input) throws IOException {
		writeln("2");
		writeln(input);
		writeln("END");

		String line;
		while (!(line = readln()).equals("INTEGRAL")) {
			if (line.equals("NOT FOUND IN ELECTRON LIST")) throw new IOException();
		}
		readln();
		nextToken();
		slaterValues = new double[2];
		slaterValues[0] = tokenizer.nval;
		readln();
		nextToken();
		slaterValues[1] = tokenizer.nval;
	}

	public void getSOParameter(String input) throws IOException {
		writeln("4");
		writeln(input);
		writeln("END");

		String line;
		while (!(line = readln()).equals("SPIN-ORBIT PARAMETER FOR")) {
			if (line.equals("NOT FOUND IN ELECTRON LIST")) throw new IOException();
		}
		readln();
		nextToken();
		soValues = new double[2];
		soValues[0] = tokenizer.nval;
		readln();
		nextToken();
		soValues[1] = tokenizer.nval;
	}

	protected String readOrbital(StreamTokenizer tokenizer) throws IOException {
		String str; 
		do {
			nextToken();
		} while (!(tokenizer.ttype == StreamTokenizer.TT_NUMBER));
		str = String.valueOf(toInt(tokenizer.nval));
		do {
			nextToken();
		} while (!(tokenizer.ttype == StreamTokenizer.TT_WORD));
		return str += tokenizer.sval;
	}

	protected int toInt(double d) {
		return new Double(d).intValue();
	}
}
